#!r6rs
(import (rnrs base (6))
        (rnrs lists(6))
        (rnrs io simple (6))
        (rnrs sorting (6))
        (rnrs eval (6)) )

;;;
;;; This program is a collection of functions that run on the sparse vector abstract data type. A sparse vector is a
;;; vector in which many of the values are 0, and such are not needed. The sparse vector representation that this program
;;; uses the format of index followed by the value at the index. Ex ((1 2) (5 2)) is the equivalent of <2, 0, 0, 0, 2>.
;;; This program has functions for 'vector sum', 'Euclidean Inner Product', 'Scalar Multiple', 'Euclidean Norm', as well
;;; as 'Euclidean Distance'. Each is explained in more detail in the documentation for each individual function.
;;;
;;; Author: Alex Dobson, University Of Alberta Augustana Campus, AUCSC 370
;;;

;; This function converts a list given in normal list format (ie (1 2 3 4 5 6) ) into the format used by all of the functions
;; for a sparse vector (ie ((1 2) (3 4) (5 6)) ).
;;
;; INPUT: l - the input list that then gets converted to the format that each function uses
;; OUPUT: returns a list in the form of a sparse vector used by the rest of the functions in this program
;;
(define initvector
  (lambda (l)
    (if (null? l) '()
        (cons (list (car l) (cadr l)) 
              (initvector (cddr l))
        )
    )
  )
)

;; This function performs a 'Vector Sum' operation on two sparse vectors. This is defined as Tr = Ur + Vr for each value
;; in U and V. The function returns a list in the form of a sparse vector.
;;
;; INPUT: vector1 - the first vector to be summed
;;        vector2 - the second vector to be summed
;; OUTPUT: a list in the form of a sparse vector of the vector sum of vector1 and vector2
;;
(define vectorsum
  (lambda (vector1 vector2) 
    (cond
      ((null? vector1) vector2)
      ((null? vector2) vector1)
      ((sameIndex? (car vector1) (car vector2)) (cons (addPairs (car vector1) (car vector2)) (vectorsum (cdr vector1) (cdr vector2))) )
      ((> (caar vector1) (caar vector2)) (cons (car vector2) (vectorsum vector1 (cdr vector2)) ) )
      ((< (caar vector1) (caar vector2)) (cons (car vector1) (vectorsum (cdr vector1) vector2) ) )
      (#t (display "SOMETHING WENT WRONG"))
    )
  )
)

;; This function is a helper function used by the vectorsum function that adds two pairs from the vectors. The pairs are
;; defined as (1 2) (1 3) in this case, where 1 is the index and 2 & 3 are the values in those indices. This functions purpose
;; is to free up space in the vectorsum function and make it easier to read.
;;
;; INPUT: pair1 - the first pair to be added
;;        pair2 - the second pair to be added
;; OUTPUT: a pair with the same index as pair1 and pair2 but with the sum of the two values
;;              ie addPairs (1 2) (1 3) ===> (1 5)
;;
(define addPairs
  (lambda (pair1 pair2)
    (list (car pair1) (+ (cadr pair1) (cadr pair2)))
  )
)

;; This function performs an inner product operation on two sparse vectors. An inner product on two vectors is defined as the
;; sum of all pairwise products of two vectors. The function returns a single number that is the inner product.
;;
;; INPUT: vector1 - the first vector to have the inner product operation performed
;;        vector2 - the second vector to have the inner product performed
;; OUTPUT: a single number that is the sum of the pairwise products of the two vectors passed as input
;;
(define innerproduct
  (lambda (vector1 vector2)
    (cond
      ((null? vector1) 0)
      ((null? vector2) 0)
      ((sameIndex? (car vector1) (car vector2)) (+ (* (cadar vector1) (cadar vector2)) (innerproduct (cdr vector1) (cdr vector2))) )
      ((> (caar vector1) (caar vector2)) (+ 0 (innerproduct vector1 (cdr vector2))) )
      ((< (caar vector1) (caar vector2)) (+ 0 (innerproduct (cdr vector1) vector2)) )
      (#t (display "SOMETHING WENT WRONG"))
    )
  )
)

;; This function performs a scalar multiplication on a vector passed as input with a scalar value. The scalar multiple operation
;; is defined as Tr = k * Vr. The function returns a list in the form of a sparse vector.
;;
;; INPUT: scalar - the scalar value to be multiplied into the vector
;;        vector - the vector to have the scalar applied to it
;; OUTPUT: a list in the form of a sparse vector that is the result of the scalar multiplication of scalar into the vector
;;
(define scalarmultiple
  (lambda (scalar vector)
    (cond
      ((null? vector) '())
      (#t (cons (scalarMultiplyPair scalar (car vector)) (scalarmultiple scalar (cdr vector))) )
    )
  )
)

;; This function is a helper function for the scalarmultiple function. It performs the multiplication of the scalar into
;; a pair which again, is defined as just the index and the value (ie (1 2) ).
;;
;; INPUT: scalar - the scalar to be multiplied into the pair
;;        pair - the pair which scalar will be multiplied into (only the value is multiplied by scalar)
;; OUTPUT: a pair with the same index as the pair passed in but with the value multiplied by scalar
;;
(define scalarMultiplyPair
  (lambda (scalar pair)
    (list (car pair) (* scalar (cadr pair)))
  )
)

;; This function performs a vector norm operation on a single vector. This is performed by applying the inner product on the 
;; vector with itself, followed by a square root of the result of the inner product.
;;
;; INPUT: vector - the vector to have the norm found
;; OUTPUT: the norm, which is a single number
;;
(define norm
  (lambda (vector)
    (sqrt (innerproduct vector vector))
  )
)

;; This funtion a vector distance operation on two vectors. This is performed finding the absolute distance between each 
;; pairwise numbers, taking the sum of those pairs and square rooting them. This function only performs the square root
;; function since it is to be performed last on the result of all the pairwise additions. The pairwise subtrations, squaring,
;; and the summation of those additions are all done by a helper function
;;
;; INPUT: vector1 - the first vector
;;        vector2 - the second vector
;; OUTPUT: the distance between vector1 and vector2 which is a number
;;
(define distance
  (lambda (vector1 vector2)
    (sqrt (distanceHelper vector1 vector2))
  )
)

;; This function is a helper function used by the distance function. This function performs the pairwise subtractions, squaring,
;; and then summation of those results. This is then passed back to the main distance funciton to be square rooted.
;;
;; INPUT: vector1 - the first vector
;;        vector2 - the second vector
;; OUTPUT: the sum of the subtrations of the pairwise products that are then squared (ie ((V1 - R1)^2 + (V2 - R2)^2 +...) )
;;
(define distanceHelper
  (lambda (vector1 vector2) (display "VECTOR 1: ") (display vector1) (display " ") (display "VECTOR 2: ") (display vector2) (newline)
    (cond
      ((and (null? vector1) (null? vector2)) 0)
      ((null? vector1) (+ (square (cadar vector2)) (distanceHelper vector1 (cdr vector2))) )
      ((null? vector2) (+ (square (cadar vector1)) (distanceHelper (cdr vector1) vector2)) )
      ((sameIndex? (car vector1) (car vector2)) 
            (+ (square (- (cadar vector1) (cadar vector2))) (distanceHelper (cdr vector1) (cdr vector2))) )
      ((> (caar vector1) (caar vector2)) (+ (square (cadar vector1)) (distanceHelper (cdr vector1) vector2)) )
      ((< (caar vector1) (caar vector2)) (+ (square (cadar vector2)) (distanceHelper vector1 (cdr vector2))) )
      (#t (display "SOMETHING WENT WRONG")(newline))
    )
  )
)

;; This function is a utility function that performs the square of a number
;;
;; INPUT: num - the number to be squared
;; OUPUT: the square of num
;;
(define square
  (lambda (num)
    (* num num)
  )
)

;; This is a utility function used to write sparse vectors in a good formatted output. The general layout of
;; the output is <I, J> <F, Y> ect... The function performs this task by taking the sparse vector and printing
;; each pair of index and value, and then calling itself recursively with the rest of the vector. When there 
;; are no more pairs to be printed, the function finishes by printing a newline character.
;;
(define writevector
  (lambda (vector)
    (cond
      ((null? vector) (newline))
      (#t (display "<") (display (caar vector)) (display ", ") (display (cadar vector)) (display "> ") (writevector (cdr vector)))
    )
  )
)

;; A utility function that checks if two pairs of index and value have the same index. This is used in order to
;; detemined if pairwise products or summation should take place in the other functions.
;;
;; INPUT: pair1 - the first pair
;;        pair2 - the second pair
;; OUTPUT: #t if the pairs share an index, #f if they do not share an index
;;
(define sameIndex?
  (lambda (pair1 pair2)
    (cond
      ((null? pair1) #f)
      ((null? pair2) #f)
      ((eqv? (car pair1) (car pair2)) #t)
      (#t #f)
    )
  )
)